------------------------------------------------------------------------
--
--                         INDEXED_IO.DEBUG_G
--
--      $Id: indexed_io-debug_g.adb,v 1.10 1999/10/09 21:27:37 matt Exp $
--
-- Copyright (C) 1999 Matthew J Heaney <matthew_heaney@acm.org>
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or (at
-- your option) any later version.
--
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
-- USA
--
-- As a special exception, if other files instantiate this generic unit,
-- or you link this unit with other files to produce an executable, this
-- unit does not by itself cause the resulting executable to be covered
-- by the GNU General Public License.  This exception does not however
-- invalidate any other reasons why the executable file might be covered
-- by the GNU Public License.
--
------------------------------------------------------------------------

with Ada.Text_IO;

package body Indexed_IO.Debug_G is

   use Stream_IO;

   procedure Put
     (File : in File_Type) is

      procedure Put_Items
        (Page_File_Index : in Ada.Streams.Stream_IO.Count) is

         Page : Page_Type;
      begin
         if Page_File_Index = 0 then
            return;
         end if;

         Read (File.File, Page, Page_File_Index);
         Put_Items (Page.Children (0));

         for Index in 1 .. Page.Length loop
            Ada.Text_IO.Put (Get_Image (Page.Items (Index)));
            Put_Items (Page.Children (Index));
         end loop;
      end Put_Items;

   begin -- Put

      if File.Root_Page.Length = 0 then
         Ada.Text_IO.Put_Line ("<empty>");
      else
         Set_Mode (File.Handle.File.File, In_File);
         Put_Items (File.Root_Page_File_Index);
         Ada.Text_IO.New_Line;
      end if;

   end Put;



   procedure Put_Tree
     (File : in File_Type) is

      procedure Put_Items
        (Page_File_Index : in Count;
         Level           : in Natural) is

         Page : Page_Type;
      begin
         if Page_File_Index = 0 then
            return;
         end if;

         Read (File.File, Page, Page_File_Index);

         for I in 1 .. 3 * Level loop
            Ada.Text_IO.Put (' ');
         end loop;

         for I in 1 .. Page.Length loop
            Ada.Text_IO.Put (Get_Image (Page.Items (I)));
         end loop;

         Ada.Text_IO.New_Line;

         for I in 0 .. Page.Length loop
            Put_Items (Page.Children (I), Level + 1);
         end loop;
      end Put_Items;

   begin -- Put_Tree

      if File.Root_Page.Length = 0 then
         Ada.Text_IO.Put_Line ("<empty>");
      else
         Set_Mode (File.Handle.File.File, In_File);
         Put_Items (File.Root_Page_File_Index, Level => 0);
      end if;

   end Put_Tree;


   function Is_In
     (File : File_Type;
      Key  : Key_Type) return Boolean is

      Index : Natural;
      Found : Boolean;

      File_Index : Stream_IO.Count := File.Root_Page_File_Index;
      Page       : Page_Type;

   begin

      Set_Mode (File.Handle.File.File, In_File);

      while File_Index /= 0 loop

         Read (File.File, Page, From => File_Index);

         Search
           (Items => Page.Items (1 .. Page.Length),
            Key   => Key,
            Index => Index,
            Found => Found);

         if Found then
            return True;
         end if;

         File_Index := Page.Children (Index);

      end loop;

      return False;

   end Is_In;


   function Get_Number_Of_Items
     (File : File_Type) return Natural is

      Number_Of_Items : Natural := 0;

      procedure Get_Length
        (Page_File_Index : in Stream_IO.Count) is

         Page : Page_Type;
      begin
         if Page_File_Index = 0 then
            return;
         end if;

         Read (File.File, Page, Page_File_Index);
         Number_Of_Items := Number_Of_Items + Page.Length;

         for I in 0 .. Page.Length loop
            Get_Length (Page.Children (I));
         end loop;
      end Get_Length;

   begin
      Set_Mode (File.Handle.File.File, In_File);
      Get_Length (File.Root_Page_File_Index);
      return Number_Of_Items;
   end;


--    function Is_Correctly_Ordered
--      (File : File_Type) return Boolean is

--       Key : Key_Type := Key_Zero;
--       OK  : Boolean := True;

--       procedure Get_Keys
--         (Page_File_Index : in Stream_IO.Count) is

--          Page : Page_Type;
--       begin
--          if Page_File_Index = 0 then
--             return;
--          end if;

--          Read (File.File, Page, Page_File_Index);
--          Get_Keys (Page.Children (0));

--          if not OK then
--             return;
--          end if;

--          for Index in 1 .. Page.Length loop

--             OK := Page.Keys (Index) > Key;

--             exit when not OK;

--             Key := Page.Keys (Index);
--             Get_Keys (Page.Children (Index));

--          end loop;
--       end Get_Keys;

--    begin

--       Set_Mode (File.Handle.File.File, In_File);
--       Get_Keys (File.Root_Page_File_Index);

--       return OK;

--    end Is_Correctly_Ordered;


   procedure Get
     (File  : in     File_Type;
      Items :    out Item_Array;
      Last  :    out Natural) is

      procedure Get_Items
        (Page_File_Index : in Stream_IO.Count) is

         Page : Page_Type;
      begin
         if Page_File_Index = 0 then
            return;
         end if;

         Read (File.File, Page, Page_File_Index);

         Get_Items (Page.Children (0));

         for Index in 1 .. Page.Length loop

            Last := Last + 1;
            Items (Last) := Page.Items (Index);
            Get_Items (Page.Children (Index));

         end loop;
      end Get_Items;

   begin

      Set_Mode (File.Handle.File.File, In_File);

      Last := Items'First - 1;

      Get_Items (File.Root_Page_File_Index);

   end Get;


   function Is_Empty
     (File : File_Type) return Boolean is
   begin
      return File.Root_Page.Length = 0;
   end;



   package Node_Management is

      function New_Node return Node_Access;

      procedure Free_Node (Node : in out Node_Access);

   end Node_Management;
   use Node_Management;


   procedure Free_Tree
     (Node : in out Node_Access) is
   begin

      if Node = null then
         return;
      end if;

      for Index in 0 .. Node.Length loop
         Free_Tree (Node.Children (Index));
      end loop;

      Free_Node (Node);

   end Free_Tree;


   procedure Copy
     (From : in     File_Type;
      To   : in out Tree_Type) is

      function Copy
        (Page_File_Index : in Ada.Streams.Stream_IO.Count)
        return Node_Access is

         Page : Page_Type;
         Node : Node_Access;
      begin
         if Page_File_Index = 0 then
            return null;
         end if;

         Read (From.File, Page, Page_File_Index);

         Node := New_Node;

         Node.Length := Page.Length;
         Node.Items (1 .. Node.Length) := Page.Items (1 .. Page.Length);

         for Index in 0 .. Page.Length loop
            Node.Children (Index) := Copy (Page.Children (Index));
         end loop;

         return Node;
      end Copy;

   begin

      Free_Tree (To.Root);

      Set_Mode (From.Handle.File.File, In_File);
      To.Root := Copy (From.Root_Page_File_Index);

   end Copy;


   procedure Put
     (Tree : in Tree_Type) is

      procedure Put_Items
        (Node  : in Node_Access;
         Level : in Natural) is
      begin
         if Node = null then
            return;
         end if;

         for I in 1 .. 3 * Level loop
            Ada.Text_IO.Put (' ');
         end loop;

         for I in 1 .. Node.Length loop
            Ada.Text_IO.Put (Get_Image (Node.Items (I)));
         end loop;

         Ada.Text_IO.New_Line;

         for I in 0 .. Node.Length loop
            Put_Items (Node.Children (I), Level + 1);
         end loop;
      end Put_Items;

   begin -- Put

      if Tree.Root = null then
         Ada.Text_IO.Put_Line ("<empty>");
      else
         Put_Items (Tree.Root, Level => 0);
      end if;

   end Put;



   package body Node_Management is

      Free_List : Node_Access;


      function New_Node return Node_Access is

         Node : Node_Access;
      begin
         if Free_List = null then
            Node := new Node_Type;
         else
            Node := Free_List;
            Free_List := Free_List.Children (0);
            Node.Children (0) := null;
         end if;

         return Node;
      end New_Node;


      procedure Free_Node (Node : in out Node_Access) is
      begin

         if Node /= null then

            Node.Children (0) := Free_List;
            Free_List := Node;

            Node := null;

         end if;

      end Free_Node;

   end Node_Management;



end Indexed_IO.Debug_G;







