------------------------------------------------------------------------
--
--                               INDEXED_IO
--
--      $Id: indexed_io.adb,v 1.38.1.54 1999/10/22 20:02:32 matt Exp $
--
-- Copyright (C) 1999 Matthew J Heaney <matthew_heaney@acm.org>
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or (at
-- your option) any later version.
--
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
-- USA
--
-- As a special exception, if other files instantiate this generic unit,
-- or you link this unit with other files to produce an executable, this
-- unit does not by itself cause the resulting executable to be covered
-- by the GNU General Public License.  This exception does not however
-- invalidate any other reasons why the executable file might be covered
-- by the GNU Public License.
--
------------------------------------------------------------------------

with Ada.Unchecked_Conversion;

package body Indexed_IO is

   -- The indexed file is implemented using a B tree.  My primary
   -- references were:
   --
   -- File Structures, 2nd ed
   -- Michael J. Folk, Bill Zoellick
   -- Addison-Wesley 1992
   --
   -- Organization and Maintenance of Large Ordered Indexes
   -- R. Bayer, E. McCreight
   -- Acta Informatica, Vol 1 (1972), p 173-189
   --
   -- The Ubiquitous B-Tree
   -- Douglas Comer
   -- Computing Surveys, Vol 11, No 2 (June 1979), p. 121-137


   procedure Insert_In_Tree
     (File                 : in out Stream_IO.File_Type;
      Root_Page_File_Index : in     Stream_IO.Positive_Count;
      Root_Page            : in out Page_Type;
      Insertion_Item       : in     Item_Type;
      Promotion            :    out Boolean;
      Promoted_Item        : in out Item_Type;
      Promoted_Right_Child : in out Stream_IO.Count);

   procedure Remove_From_Tree
     (File            : in out Stream_IO.File_Type;
      Page_File_Index : in     Stream_IO.Positive_Count;
      Page            : in out Page_Type;
      Key             : in     Key_Type;
      Underflow       :    out Boolean);


   use Stream_IO;


   --------------
   -- Page I/O --
   --------------

   procedure Write
     (File : in Stream_IO.File_Type;
      Page : in Page_Type;
      To   : in Stream_IO.Positive_Count) is

      Size : constant Stream_Element_Count :=
        Page'Size / Stream_Element_Array'Component_Size;

      subtype Page_As_Stream_Element_Array is
        Stream_Element_Array (1 .. Size);

      pragma Assert (Page'Size = Page_As_Stream_Element_Array'Size);

      function To_Stream_Element_Array is
        new Ada.Unchecked_Conversion
        (Page_Type,
         Page_As_Stream_Element_Array);
   begin
      Write (File, To_Stream_Element_Array (Page), To);
   end;


   procedure Append
     (File       : in out Stream_IO.File_Type;
      Page       : in     Page_Type;
      File_Index :    out Stream_IO.Positive_Count) is

      Size : constant Stream_Element_Count :=
        Page'Size / Stream_Element_Array'Component_Size;

      subtype Page_As_Stream_Element_Array is
        Stream_Element_Array (1 .. Size);

      pragma Assert (Page'Size = Page_As_Stream_Element_Array'Size);

      function To_Stream_Element_Array is
        new Ada.Unchecked_Conversion
        (Page_Type,
         Page_As_Stream_Element_Array);
   begin
      Set_Mode (File, Append_File);
      File_Index := Index (File);
      Write (File, To_Stream_Element_Array (Page));
   end;


   procedure Read
     (File : in     Stream_IO.File_Type;
      Page :    out Page_Type;
      From : in     Stream_IO.Positive_Count) is

      Size : constant Stream_Element_Count :=
        Page'Size / Stream_Element_Array'Component_Size;

      subtype Page_As_Stream_Element_Array is
        Stream_Element_Array (1 .. Size);

      Page_As_Stream_Elements : Page_As_Stream_Element_Array;
      for Page_As_Stream_Elements'Address use Page'Address;

      pragma Assert (Page'Size = Page_As_Stream_Element_Array'Size);

      Last : Ada.Streams.Stream_Element_Offset;
   begin
      Read (File, Page_As_Stream_Elements, Last, From);
      pragma Assert (Last = Page_As_Stream_Elements'Last);
   end;



   procedure Create
     (File : in out File_Type;
      Name : in     String := "") is
   begin

      Create (File.File, Out_File, Name);

      -- Allocate disk space for root page file index, which is always
      -- placed at the start of the file.

      Count'Write (Stream (File.File), Count'(0));


      -- Allocate disk space for (empty) root page, and cache its file
      -- index.

      File.Root_Page.Length := 0;
      File.Root_Page.Children (0) := 0;

      Append (File.File, File.Root_Page, File.Root_Page_File_Index);


      -- Now save the actual value of the root page file index.

      Reset (File.File);

      Count'Write (Stream (File.File), File.Root_Page_File_Index);

   end Create;


   procedure Open
     (File : in out File_Type;
      Name : in     String) is
   begin

      Open (File.File, In_File, Name);

      Count'Read (Stream (File.File), File.Root_Page_File_Index);

      Read (File.File, File.Root_Page, File.Root_Page_File_Index);

   end Open;


   procedure Close
     (File : in out File_Type) is
   begin
      Stream_IO.Close (File.File);
   end;


   function Is_Open
     (File : File_Type) return Boolean is
   begin
      return Stream_IO.Is_Open (File.File);
   end;



   procedure Read
     (File : in     File_Type;
      Key  : in     Key_Type;
      Item :    out Item_Type) is

      Index : Natural;
      Found : Boolean;

      File_Index : Stream_IO.Count;
      Page       : Page_Type;

   begin

      Set_Mode (File.Handle.File.File, In_File);

      Search
        (Items => File.Root_Page.Items (1 .. File.Root_Page.Length),
         Key   => Key,
         Index => Index,
         Found => Found);

      if Found then

         Item := File.Root_Page.Items (Index);

         return;

      end if;


      File_Index := File.Root_Page.Children (Index);

      while File_Index /= 0 loop

         Read (File.File, Page, From => File_Index);

         Search
           (Items => Page.Items (1 .. Page.Length),
            Key   => Key,
            Index => Index,
            Found => Found);

         if Found then

            Item := Page.Items (Index);

            return;

         end if;

         File_Index := Page.Children (Index);

      end loop;

      raise Item_Not_Found;

   end Read;


   procedure Select_Items
     (File : in File_Type) is

      Done : Boolean := False;

      procedure Select_Items_In_Tree
        (Page_File_Index : in Stream_IO.Count) is

         Page : Page_Type;
      begin
         if Page_File_Index = 0 then
            return;
         end if;

         Read (File.File, Page, From => Page_File_Index);

         Select_Items_In_Tree (Page.Children (0));

         if Done then
            return;
         end if;

         for I in Positive range 1 .. Page.Length loop

            Process (Page.Items (I), Done);
            exit when Done;

            Select_Items_In_Tree (Page.Children (I));
            exit when Done;

         end loop;
      end Select_Items_In_Tree;

   begin -- Select_Items

      Set_Mode (File.Handle.File.File, In_File);

      Select_Items_In_Tree (File.Root_Page.Children (0));

      if Done then
         return;
      end if;

      for I in 1 .. File.Root_Page.Length loop

         Process (File.Root_Page.Items (I), Done);
         exit when Done;

         Select_Items_In_Tree (File.Root_Page.Children (I));
         exit when Done;

      end loop;

   end Select_Items;


   procedure Select_Items_Starting_At_Key
     (File : in File_Type;
      Key  : in Key_Type) is

      Done : Boolean := False;


      procedure Select_Items_In_Tree
        (Page_File_Index : in Stream_IO.Count) is

         Page : Page_Type;

      begin

         if Page_File_Index = 0 then
            return;
         end if;

         Read (File.File, Page, From => Page_File_Index);

         Select_Items_In_Tree (Page.Children (0));

         if Done then
            return;
         end if;

         for I in 1 .. Page.Length loop

            Process (Page.Items (I), Done);

            if Done then
               return;
            end if;

            Select_Items_In_Tree (Page.Children (I));

            if Done then
               return;
            end if;

         end loop;

      end Select_Items_In_Tree;


      procedure Search_For_Key
        (Page_File_Index : in Stream_IO.Count) is

         Index : Natural;
         Found : Boolean;

         Page : Page_Type;

      begin

         if Page_File_Index = 0 then
            return;
         end if;

         Read (File.File, Page, From => Page_File_Index);

         Search
           (Items => Page.Items (1 .. Page.Length),
            Key   => Key,
            Index => Index,
            Found => Found);

         if Found then

            Process (Page.Items (Index), Done);

            if Done then
               return;
            end if;

            Select_Items_In_Tree (Page.Children (Index));

         else

            Search_For_Key (Page.Children (Index));

         end if;

         if Done then
            return;
         end if;

         for I in Index + 1 .. Page.Length loop

            Process (Page.Items (I), Done);

            if Done then
               return;
            end if;

            Select_Items_In_Tree (Page.Children (I));

            if Done then
               return;
            end if;

         end loop;

      end Search_For_Key;


      Index : Natural;
      Found : Boolean;

   begin -- Select_Items_Starting_At_Key

      Set_Mode (File.Handle.File.File, In_File);

      Search
        (Items => File.Root_Page.Items (1 .. File.Root_Page.Length),
         Key   => Key,
         Index => Index,
         Found => Found);

      if Found then

         Process (File.Root_Page.Items (Index), Done);

         if Done then
            return;
         end if;

         Select_Items_In_Tree (File.Root_Page.Children (Index));

      else

         Search_For_Key (File.Root_Page.Children (Index));

      end if;

      if Done then
         return;
      end if;

      for I in Index + 1 .. File.Root_Page.Length loop

        Process (File.Root_Page.Items (I), Done);

        if Done then
           return;
        end if;

        Select_Items_In_Tree (File.Root_Page.Children (I));

        if Done then
           return;
        end if;

      end loop;

   end Select_Items_Starting_At_Key;




   procedure Do_Write
     (File : in out File_Type;
      Item : in     Item_Type) is

      Promotion : Boolean;

      Item_Promoted_From_Below        : Item_Type;
      Right_Child_Promoted_From_Below : Stream_IO.Count;

   begin

      Set_Mode (File.File, In_File);

      Insert_In_Tree
        (File                 => File.File,
         Root_Page_File_Index => File.Root_Page_File_Index,
         Root_Page            => File.Root_Page,
         Insertion_Item       => Item,
         Promotion            => Promotion,
         Promoted_Item        => Item_Promoted_From_Below,
         Promoted_Right_Child => Right_Child_Promoted_From_Below);

      if not Promotion then
         return;
      end if;


      -- There's been a promotion, up from the root page, which means the
      -- height of the tree has just increased.  Create a new root page, to
      -- store the promoted item.

      File.Root_Page.Items (1) := Item_Promoted_From_Below;

      File.Root_Page.Children (0) := File.Root_Page_File_Index;
      File.Root_Page.Children (1) := Right_Child_Promoted_From_Below;

      File.Root_Page.Length := 1;


      -- Write the new root page to disk, and record its file index.
      Append (File.File, File.Root_Page, File.Root_Page_File_Index);

      -- Write the file index of the new root page to disk.
      Reset (File.File);
      Count'Write (Stream (File.File), File.Root_Page_File_Index);

      Flush (File.File);

   end Do_Write;


   procedure Write
     (File : in File_Type;
      Item : in Item_Type) is
   begin
      Do_Write (File.Handle.File.all, Item);
   end;


   procedure Do_Remove
     (File : in out File_Type;
      Key  : in     Key_Type) is

      Underflow : Boolean;

   begin

      Set_Mode (File.File, In_File);

      Remove_From_Tree
        (File            => File.File,
         Page_File_Index => File.Root_Page_File_Index,
         Page            => File.Root_Page,
         Key             => Key,
         Underflow       => Underflow);

      if Underflow then

         if File.Root_Page.Length = 0            -- root page is empty
           and File.Root_Page.Children (0) /= 0  -- and not a leaf
         then

            -- Reduce the height of the tree.

            File.Root_Page_File_Index := File.Root_Page.Children (0);
            --
            -- This throws away whatever page was designated by
            -- Root_Page_File_Index, leaving a "hole" in the file.

            Reset (File.File, Out_File);

            Count'Write (Stream (File.File), File.Root_Page_File_Index);


            -- Cache page that's the new root.

            Set_Mode (File.File, In_File);

            Read (File.File, File.Root_Page, File.Root_Page_File_Index);

         else

            -- Just save the root page, without reducing the height.

            Set_Mode (File.File, Out_File);

            Write (File.File, File.Root_Page, File.Root_Page_File_Index);

         end if;

      end if;


      Set_Mode (File.File, Out_File);

      Flush (File.File);

   end Do_Remove;


   procedure Remove
     (File : in File_Type;
      Key  : in Key_Type) is
   begin
      Do_Remove (File.Handle.File.all, Key);
   end;




   ---------------
   -- Insertion --
   ---------------

   procedure Insert_In_Split_Page
     (File                  : in out Stream_IO.File_Type;
      Page_File_Index       : in     Stream_IO.Positive_Count;
      Page                  : in out Page_Type;
      Index                 : in     Natural;
      Insertion_Item        : in     Item_Type;
      Insertion_Right_Child : in     Stream_IO.Count;
      Promoted_Item         :    out Item_Type;
      Promoted_Right_Child  :    out Stream_IO.Positive_Count) is

      New_Right_Sibling : Page_Type;
      -- This is the page whose file index is Promoted_Right_Child.

   begin

      if Index < Order then

         Promoted_Item := Page.Items (Order);


         New_Right_Sibling.Items (1 .. Order) :=
           Page.Items (Order + 1 .. Page_Size);

         New_Right_Sibling.Children (0 .. Order) :=
           Page.Children (Order .. Page_Size);


         Page.Items (Index + 1 .. Order) :=
           Insertion_Item &
           Page.Items (Index + 1 .. Order - 1);

         Page.Children (Index + 1 .. Order) :=
           Insertion_Right_Child &
           Page.Children (Index + 1 .. Order - 1);


      elsif Index = Order then

         Promoted_Item := Insertion_Item;


         New_Right_Sibling.Items (1 .. Order) :=
           Page.Items (Order + 1 .. Page_Size);

         New_Right_Sibling.Children (0 .. Order) :=
           Insertion_Right_Child &
           Page.Children (Order + 1 .. Page_Size);


      else

         Promoted_Item := Page.Items (Order + 1);

         New_Right_Sibling.Items (1 .. Order) :=
           Page.Items (Order + 2 .. Index) &
           Insertion_Item &
           Page.Items (Index + 1 .. Page_Size);

         New_Right_Sibling.Children (0 .. Order) :=
           Page.Children (Order + 1 .. Index) &
           Insertion_Right_Child &
           Page.Children (Index + 1 .. Page_Size);


      end if;

      Page.Length := Order;
      New_Right_Sibling.Length := Order;

      Append (File, New_Right_Sibling, Promoted_Right_Child);
      Write (File, Page, Page_File_Index);

   end Insert_In_Split_Page;


   procedure Insert_In_Page
     (File                  : in out Stream_IO.File_Type;
      Page_File_Index       : in     Stream_IO.Positive_Count;
      Page                  : in out Page_Type;
      Index                 : in     Natural;
      Insertion_Item        : in     Item_Type;
      Insertion_Right_Child : in     Stream_IO.Count) is

   begin

      Page.Items (Index + 1 .. Page.Length + 1) :=
        Insertion_Item &
        Page.Items (Index + 1 .. Page.Length);

      Page.Children (Index + 1 .. Page.Length + 1) :=
        Insertion_Right_Child &
        Page.Children (Index + 1 .. Page.Length);

      Page.Length := Page.Length + 1;


      Set_Mode (File, Out_File);
      Write (File, Page, Page_File_Index);

   end Insert_In_Page;


   procedure Insert_In_Page_And_Right_Sibling
     (File                  : in out Stream_IO.File_Type;
      Parent_File_Index     : in     Stream_IO.Positive_Count;
      Parent                : in out Page_Type;
      Left_Index            : in     Child_Range;
      Left                  : in out Page_Type;
      Right                 : in out Page_Type;
      Item_Index            : in     Natural;
      Insertion_Item        : in     Item_Type;
      Insertion_Right_Child : in     Stream_IO.Count) is

      pragma Assert (Left.Length = Page_Size);

      Parent_Item : Item_Type renames
        Parent.Items (Left_Index + 1);

      Empty_Slots_Count : constant Positive :=
        Page_Size - Right.Length;
      --
      -- The total number of free slots in the right sibling.

      Move_Count : constant Positive :=
        Integer'Max (1, Empty_Slots_Count / 2);
      --
      -- The number of items we can move (from the left sibling, the
      -- parent, and the insertion item itself) into the right sibling.

      Parent_Item_Index : constant Positive :=
        Page_Size - Move_Count + 1;
      --
      -- The index of the item that moves up, into the parent.

      Left_File_Index : Stream_IO.Count renames
        Parent.Children (Left_Index);

      Right_File_Index : Stream_IO.Count renames
        Parent.Children (Left_Index + 1);

   begin

      if Item_Index < Parent_Item_Index then

         -- The insertion item goes in the left sibling.

         Right.Items (1 .. Right.Length + Move_Count) :=
           Left.Items (Parent_Item_Index + 1 .. Page_Size) &
           Parent_Item &
           Right.Items (1 .. Right.Length);

         Right.Children (0 .. Right.Length + Move_Count) :=
           Left.Children (Parent_Item_Index .. Page_Size) &
           Right.Children (0 .. Right.Length);

         Parent_Item := Left.Items (Parent_Item_Index);

         Left.Items (Item_Index + 1 .. Parent_Item_Index) :=
           Insertion_Item &
           Left.Items (Item_Index + 1 .. Parent_Item_Index - 1);

         Left.Children (Item_Index + 1 .. Parent_Item_Index) :=
           Insertion_Right_Child &
           Left.Children (Item_Index + 1 .. Parent_Item_Index - 1);

      elsif Item_Index = Parent_Item_Index then

         -- The insertion item moves up into the parent.

         Right.Items (1 .. Right.Length + Move_Count) :=
           Left.Items (Item_Index + 1 .. Page_Size) &
           Parent_Item &
           Right.Items (1 .. Right.Length);

         Right.Children (0 .. Right.Length + Move_Count) :=
           Insertion_Right_Child &
           Left.Children (Item_Index + 1 .. Page_Size) &
           Right.Children (0 .. Right.Length);

         Parent_Item := Insertion_Item;

      else

         -- The insertion item goes in the right sibling.

         Right.Items (1 .. Right.Length + Move_Count) :=
           Left.Items (Parent_Item_Index + 2 .. Item_Index) &
           Insertion_Item &
           Left.Items (Item_Index + 1 .. Page_Size) &
           Parent_Item &
           Right.Items (1 .. Right.Length);

         Right.Children (0 .. Right.Length + Move_Count) :=
           Left.Children (Parent_Item_Index + 1 .. Item_Index) &
           Insertion_Right_Child &
           Left.Children (Item_Index + 1 .. Page_Size) &
           Right.Children (0 .. Right.Length);

         Parent_Item := Left.Items (Parent_Item_Index + 1);

      end if;

      Left.Length := Left.Length - Move_Count + 1;
      Right.Length := Right.Length + Move_Count;


      Set_Mode (File, Out_File);

      Write (File, Parent, Parent_File_Index);
      Write (File, Left, Left_File_Index);
      Write (File, Right, Right_File_Index);

   end Insert_In_Page_And_Right_Sibling;


   procedure Insert_In_Page_And_Left_Sibling
     (File                  : in out Stream_IO.File_Type;
      Parent_File_Index     : in     Stream_IO.Positive_Count;
      Parent                : in out Page_Type;
      Left                  : in out Page_Type;
      Right_Index           : in     Natural;
      Right                 : in out Page_Type;
      Item_Index            : in     Natural;
      Insertion_Item        : in     Item_Type;
      Insertion_Right_Child : in     Stream_IO.Count) is

      pragma Assert (Right.Length = Page_Size);

      Parent_Item : Item_Type renames
        Parent.Items (Right_Index);

      Empty_Slots_Count : constant Positive :=
        Page_Size - Left.Length;
      --
      -- The total number of free slots in the left sibling.

      Move_Count : constant Positive :=
        Integer'Max (1, Empty_Slots_Count / 2);
      --
      -- The number of items that actually move.

      Parent_Item_Index : constant Natural := Move_Count - 1;
      --
      -- Index position of item on right page that will move up into parent
      -- page.

      Left_File_Index : Stream_IO.Count renames
        Parent.Children (Right_Index - 1);

      Right_File_Index : Stream_IO.Count renames
        Parent.Children (Right_Index);

   begin

      if Item_Index < Parent_Item_Index then

         -- Insertion_Item goes in left sibling.

         Left.Items (Left.Length + 1 .. Left.Length + Move_Count) :=
           Parent_Item &
           Right.Items (1 .. Item_Index) &
           Insertion_Item &
           Right.Items (Item_Index + 1 .. Move_Count - 2);

         Left.Children (Left.Length + 1 .. Left.Length + Move_Count) :=
           Right.Children (0 .. Item_Index) &
           Insertion_Right_Child &
           Right.Children (Item_Index + 1 .. Move_Count - 2);

         Parent_Item := Right.Items (Move_Count - 1);

         Right.Items (1 .. Right.Length - Move_Count + 1) :=
           Right.Items (Move_Count .. Right.Length);

         Right.Children (0 .. Right.Length - Move_Count + 1) :=
           Right.Children (Move_Count - 1 .. Right.Length);

      elsif Item_Index = Parent_Item_Index then

         -- Insertion_Item goes in parent page.

         Left.Items (Left.Length + 1 .. Left.Length + Move_Count) :=
           Parent_Item &
           Right.Items (1 .. Move_Count - 1);

         Left.Children (Left.Length + 1 .. Left.Length + Move_Count) :=
           Right.Children (0 .. Move_Count - 1);

         Parent_Item := Insertion_Item;

         Right.Items (1 .. Right.Length - Move_Count + 1) :=
           Right.Items (Move_Count .. Right.Length);

         Right.Children (0 .. Right.Length - Move_Count + 1) :=
           Insertion_Right_Child &
           Right.Children (Move_Count .. Right.Length);

      else

         -- Insertion_Item goes in right sibling.

         Left.Items (Left.Length + 1 .. Left.Length + Move_Count) :=
           Parent_Item &
           Right.Items (1 .. Move_Count - 1);

         Left.Children (Left.Length + 1 .. Left.Length + Move_Count) :=
           Right.Children (0 .. Move_Count - 1);

         Parent_Item := Right.Items (Move_Count);

         Right.Items (1 .. Right.Length - Move_Count + 1) :=
           Right.Items (Move_Count + 1 .. Item_Index) &
           Insertion_Item &
           Right.Items (Item_Index + 1 .. Right.Length);

         Right.Children (0 .. Right.Length - Move_Count + 1) :=
           Right.Children (Move_Count .. Item_Index) &
           Insertion_Right_Child &
           Right.Children (Item_Index + 1 .. Right.Length);

      end if;

      Left.Length := Left.Length + Move_Count;
      Right.Length := Right.Length - Move_Count + 1;


      Set_Mode (File, Out_File);

      Write (File, Parent, Parent_File_Index);
      Write (File, Left, Left_File_Index);
      Write (File, Right, Right_File_Index);

   end Insert_In_Page_And_Left_Sibling;



   procedure Insert_In_Subtree
     (File                 : in out Stream_IO.File_Type;
      Parent_File_Index    : in     Stream_IO.Positive_Count;
      Parent               : in out Page_Type;
      Page_Index           : in     Child_Range;
      Insertion_Item       : in     Item_Type;
      Promotion            :    out Boolean;
      Promoted_Item        : in out Item_Type;
      Promoted_Right_Child : in out Stream_IO.Count) is

      Sibling : Page_Type;

      function Get_Sibling
        (Sibling_Index : Child_Range) return Positive is
      begin
         Set_Mode (File, In_File);
         Read (File, Sibling, Parent.Children (Sibling_Index));
         return Sibling.Length;
      end;

      Page_File_Index : Stream_IO.Count renames
        Parent.Children (Page_Index);

      Page : Page_Type;

      Search_Index : Natural;
      Found        : Boolean;

      Item_Promoted_From_Below        : Item_Type;
      Right_Child_Promoted_From_Below : Stream_IO.Count;

   begin

      if Page_File_Index = 0 then

         -- We've descended all the way to the bottom, below the tree,
         -- because the item wasn't already in the tree.  Promote the item
         -- up into the leaf.

         Promotion := True;

         Promoted_Item := Insertion_Item;
         Promoted_Right_Child := 0;

         return;

      end if;


--    Set_Mode (File, In_File);
      Read (File, Page, Page_File_Index);

      Search
        (Items => Page.Items (1 .. Page.Length),
         Key   => Get_Key (Insertion_Item),
         Index => Search_Index,
         Found => Found);

      if Found then

         -- Is it an error to write the item if the key is already in the
         -- file?  I just interpret that to mean that the client wants to
         -- replace the existing item having that key.

         Page.Items (Search_Index) := Insertion_Item;

         Set_Mode (File, Out_File);
         Write (File, Page, Page_File_Index);

         Promotion := False;

         return;

      end if;


      -- The item wasn't found in this node.  Insert the item in the left
      -- subtree.

      Insert_In_Subtree
        (File                 => File,
         Parent_File_Index    => Page_File_Index,
         Parent               => Page,
         Page_Index           => Search_Index,
         Insertion_Item       => Insertion_Item,
         Promotion            => Promotion,
         Promoted_Item        => Item_Promoted_From_Below,
         Promoted_Right_Child => Right_Child_Promoted_From_Below);


      if not Promotion then

         -- The item was inserted in the (left) subtree, and no item was
         -- promoted from below.  Therefore, we're done.

         return;

      end if;


      -- The item was inserted in the child subtree, but a split occurred
      -- and another item was promoted up from below.  Insertion of the
      -- item is handled as follows:
      --
      -- 1) Insert the promoted item in this page if there's room.
      --
      -- 2) Insert the item in a sibling if there's room.
      --
      -- 3) Otherwise, split this page and promote the middle item up into
      -- our parent.
      --
      -- The idea is that we want to delay splitting a node for as long as
      -- possible.  This increases the space efficiency, and minimizes the
      -- height of the tree.

      if Page.Length < Page_Size then

         Insert_In_Page
           (File                  => File,
            Page_File_Index       => Page_File_Index,
            Page                  => Page,
            Index                 => Search_Index,
            Insertion_Item        => Item_Promoted_From_Below,
            Insertion_Right_Child => Right_Child_Promoted_From_Below);

         Promotion := False;

      elsif Page_Index = 0 then

          -- This is the left-most child of the parent, and therefore has no
          -- left sibling.  Redistribute with the right sibling if possible.

         if Get_Sibling (1) < Page_Size then

            Insert_In_Page_And_Right_Sibling
              (File                  => File,
               Parent_File_Index     => Parent_File_Index,
               Parent                => Parent,
               Left_Index            => Page_Index,
               Left                  => Page,
               Right                 => Sibling,
               Item_Index            => Search_Index,
               Insertion_Item        => Item_Promoted_From_Below,
               Insertion_Right_Child => Right_Child_Promoted_From_Below);

            Promotion := False;

         else

            Insert_In_Split_Page
              (File                  => File,
               Page_File_Index       => Page_File_Index,
               Page                  => Page,
               Index                 => Search_Index,
               Insertion_Item        => Item_Promoted_From_Below,
               Insertion_Right_Child => Right_Child_Promoted_From_Below,
               Promoted_Item         => Promoted_Item,
               Promoted_Right_Child  => Promoted_Right_Child);

          end if;


      elsif Page_Index = Parent.Length then

         -- This is the right-most child of the parent, and therefore has no
         -- right sibling.  Redistribute with the left sibling if possible.

         if Get_Sibling (Page_Index - 1) < Page_Size then

            Insert_In_Page_And_Left_Sibling
              (File                  => File,
               Parent_File_Index     => Parent_File_Index,
               Parent                => Parent,
               Left                  => Sibling,
               Right_Index           => Page_Index,
               Right                 => Page,
               Item_Index            => Search_Index,
               Insertion_Item        => Item_Promoted_From_Below,
               Insertion_Right_Child => Right_Child_Promoted_From_Below);

            Promotion := False;

         else

            Insert_In_Split_Page
              (File                  => File,
               Page_File_Index       => Page_File_Index,
               Page                  => Page,
               Index                 => Search_Index,
               Insertion_Item        => Item_Promoted_From_Below,
               Insertion_Right_Child => Right_Child_Promoted_From_Below,
               Promoted_Item         => Promoted_Item,
               Promoted_Right_Child  => Promoted_Right_Child);

         end if;

      elsif Get_Sibling (Page_Index - 1) < Page_Size then

         -- There's room in the left sibling.

         Insert_In_Page_And_Left_Sibling
           (File                  => File,
            Parent_File_Index     => Parent_File_Index,
            Parent                => Parent,
            Left                  => Sibling,
            Right_Index           => Page_Index,
            Right                 => Page,
            Item_Index            => Search_Index,
            Insertion_Item        => Item_Promoted_From_Below,
            Insertion_Right_Child => Right_Child_Promoted_From_Below);

         Promotion := False;

      elsif Get_Sibling (Page_Index + 1) < Page_Size then

         -- There's room in the right sibling.

         Insert_In_Page_And_Right_Sibling
           (File                  => File,
            Parent_File_Index     => Parent_File_Index,
            Parent                => Parent,
            Left_Index            => Page_Index,
            Left                  => Page,
            Right                 => Sibling,
            Item_Index            => Search_Index,
            Insertion_Item        => Item_Promoted_From_Below,
            Insertion_Right_Child => Right_Child_Promoted_From_Below);

         Promotion := False;

      else

         Insert_In_Split_Page
           (File                  => File,
            Page_File_Index       => Page_File_Index,
            Page                  => Page,
            Index                 => Search_Index,
            Insertion_Item        => Item_Promoted_From_Below,
            Insertion_Right_Child => Right_Child_Promoted_From_Below,
            Promoted_Item         => Promoted_Item,
            Promoted_Right_Child  => Promoted_Right_Child);

      end if;

   end Insert_In_Subtree;


   procedure Insert_In_Tree
     (File                 : in out Stream_IO.File_Type;
      Root_Page_File_Index : in     Stream_IO.Positive_Count;
      Root_Page            : in out Page_Type;
      Insertion_Item       : in     Item_Type;
      Promotion            :    out Boolean;
      Promoted_Item        : in out Item_Type;
      Promoted_Right_Child : in out Stream_IO.Count) is

      Search_Index : Natural;
      Found        : Boolean;

      Item_Promoted_From_Below        : Item_Type;
      Right_Child_Promoted_From_Below : Stream_IO.Count;

   begin

      Search
        (Items => Root_Page.Items (1 .. Root_Page.Length),
         Key   => Get_Key (Insertion_Item),
         Index => Search_Index,
         Found => Found);

      if Found then

         Root_Page.Items (Search_Index) := Insertion_Item;

         Set_Mode (File, Out_File);

         Write (File, Root_Page, Root_Page_File_Index);

         Promotion := False;

         return;

      end if;


      Insert_In_Subtree
        (File                 => File,
         Parent_File_Index    => Root_Page_File_Index,
         Parent               => Root_Page,
         Page_Index           => Search_Index,
         Insertion_Item       => Insertion_Item,
         Promotion            => Promotion,
         Promoted_Item        => Item_Promoted_From_Below,
         Promoted_Right_Child => Right_Child_Promoted_From_Below);

      if not Promotion then

         return;

      end if;


      if Root_Page.Length < Page_Size then

         Insert_In_Page
           (File                  => File,
            Page_File_Index       => Root_Page_File_Index,
            Page                  => Root_Page,
            Index                 => Search_Index,
            Insertion_Item        => Item_Promoted_From_Below,
            Insertion_Right_Child => Right_Child_Promoted_From_Below);

         Promotion := False;


      else

         Insert_In_Split_Page
           (File                  => File,
            Page_File_Index       => Root_Page_File_Index,
            Page                  => Root_Page,
            Index                 => Search_Index,
            Insertion_Item        => Item_Promoted_From_Below,
            Insertion_Right_Child => Right_Child_Promoted_From_Below,
            Promoted_Item         => Promoted_Item,
            Promoted_Right_Child  => Promoted_Right_Child);

      end if;

   end Insert_In_Tree;



   --------------
   -- Removal --
   --------------

   procedure Redistribute_Items_From_Right_To_Left
     (Parent      : in out Page_Type;
      Left        : in out Page_Type;
      Right_Index : in     Positive;
      Right       : in out Page_Type) is

      -- Description:
      --
      -- Move some extra items from the right page onto its left sibling.
      --
      -- This is a redistribution, so it modifies the parent (Parent) and
      -- both children (Left, Right).  It does NOT write the pages to disk.


      pragma Assert (Left.Length = Order - 1);

      Extra_Items_Count : constant Positive :=
        Right.Length - Order;
      --
      -- Total number of items "available" for moving from the right page,
      -- without violating constraint that there are at least Order items
      -- in a page (except root).

      Move_Count : constant Positive :=
        Integer'Max (1, Extra_Items_Count / 2);
      --
      -- Move half of what's available on the right to the left; this
      -- evenly distributes the items.

   begin

      Left.Length := Order + Move_Count - 1;

      Left.Items (Order .. Left.Length) :=
        Parent.Items (Right_Index) &
        Right.Items (1 .. Move_Count - 1);

      Left.Children (Order .. Left.Length) :=
        Right.Children (0 .. Move_Count - 1);


      Parent.Items (Right_Index) := Right.Items (Move_Count);


      Right.Items (1 .. Right.Length - Move_Count) :=
        Right.Items (Move_Count + 1 .. Right.Length);

      Right.Children (0 .. Right.Length - Move_Count) :=
        Right.Children (Move_Count .. Right.Length);

      Right.Length := Right.Length - Move_Count;

   end Redistribute_Items_From_Right_To_Left;


   procedure Redistribute_Items_From_Left_To_Right
     (Parent      : in out Page_Type;
      Left        : in out Page_Type;
      Right_Index : in     Positive;
      Right       : in out Page_Type) is

      -- Description:
      --
      -- Move some extra items from the left page onto its right sibling.
      --
      -- This is a redistribution, so it modifies the parent (Parent) and
      -- both siblings (Left, Right).  It does NOT write the pages to disk.


      pragma Assert (Right.Length = Order - 1);

      Parent_Item : Item_Type renames Parent.Items (Right_Index);

      Extra_Items_Count : constant Positive :=
        Left.Length - Order;
      --
      -- Total number of items "available" for moving from the left page,
      -- without violating constraint that there are at least Order items
      -- in a page (except root).

      Move_Count : constant Positive :=
        Integer'Max (1, Extra_Items_Count / 2);
      --
      -- Move half of what's available on the left to the right; this
      -- evenly distributes the items.

   begin

      Right.Items (1 .. Right.Length + Move_Count) :=
        Left.Items (Left.Length - Move_Count + 2 .. Left.Length) &
        Parent_Item &
        Right.Items (1 .. Right.Length);

      Right.Children (0 .. Right.Length + Move_Count) :=
        Left.Children (Left.Length - Move_Count + 1 .. Left.Length) &
        Right.Children (0 .. Right.Length);

      Right.Length := Right.Length + Move_Count;


      Parent_Item := Left.Items (Left.Length - Move_Count + 1);


      Left.Length := Left.Length - Move_Count;

   end Redistribute_Items_From_Left_To_Right;


   procedure Concatenate
     (Parent      : in out Page_Type;
      Left        : in out Page_Type;
      Right_Index : in     Positive;
      Right       : in out Page_Type) is

      -- Description:
      --
      -- Combines a pair of sibling pages into one, by demoting the
      -- Index'th item from the Parent page and copying all the items from
      -- the Right page into the Left.
      --
      -- This means that the file will now contain a "hole" where the Right
      -- page used to be.
      --
      -- The Parent page and both siblings (Right, Left) are all modified.
      -- This subprogram does NOT write the pages to disk.

   begin

      Left.Items (Left.Length + 1 .. Page_Size) :=
        Parent.Items (Right_Index) &
        Right.Items (1 .. Right.Length);

      Left.Children (Left.Length + 1 .. Page_Size) :=
        Right.Children (0 .. Right.Length);

      Left.Length := Page_Size;


      Parent.Items (Right_Index .. Parent.Length - 1) :=
        Parent.Items (Right_Index + 1 .. Parent.Length);

      Parent.Children (Right_Index .. Parent.Length - 1) :=
        Parent.Children (Right_Index + 1 .. Parent.Length);

      Parent.Length := Parent.Length - 1;

   end Concatenate;


   procedure Redistribute_Or_Concatenate_With_Right_Sibling
     (File       : in out Stream_IO.File_Type;
      Parent     : in out Page_Type;
      Left_Index : in     Natural;
      Left       : in out Page_Type;
      Underflow  :    out Boolean) is

      -- Description:
      --
      -- Repair an underflow in the Left page, by either moving some extra
      -- items from its right sibling, or by combining the siblings
      -- together to form a single page.
      --
      -- The Underflow flag indicates whether the Parent page has
      -- underflowed.
      --
      -- The Parent page and both of its children are all modified.  This
      -- subprogram writes the children to disk (as necessary), but it does
      -- NOT write the Parent page to disk.


      pragma Assert (Left.Length < Order);

      Left_File_Index : constant Stream_IO.Positive_Count :=
        Parent.Children (Left_Index);

      Right_Index : constant Positive :=
        Left_Index + 1;

      Right_File_Index : constant Stream_IO.Positive_Count :=
        Parent.Children (Right_Index);

      Right : Page_Type;

   begin

      Set_Mode (File, In_File);
      Read (File, Right, Right_File_Index);

      Set_Mode (File, Out_File);

      if Right.Length > Order then

         -- There are some extra items available in the right sibling, so
         -- move some of them onto the left page.
         --
         Redistribute_Items_From_Right_To_Left
           (Parent      => Parent,
            Left        => Left,
            Right_Index => Right_Index,
            Right       => Right);


         Underflow := False;
         -- Let the caller know that the Parent page hasn't underflowed.


         Write (File, Right, Right_File_Index);

      else

         -- There aren't any extra items in the right sibling.  So move
         -- all the items from the right into the left, and throw away the
         -- right page.
         --
         Concatenate
           (Parent      => Parent,
            Left        => Left,
            Right_Index => Right_Index,
            Right       => Right);

         -- Let the caller know whether the Parent page has underflowed.
         Underflow := Parent.Length < Order;

         --
         -- Note that we don't write the Right page to disk, because that
         -- page is gone.  All of its items were moved to the Left page as
         -- a result of the concatentation.
         --

      end if;

      Write (File, Left, Left_File_Index);

    end Redistribute_Or_Concatenate_With_Right_Sibling;



   procedure Redistribute_Or_Concatenate_With_Left_Sibling
     (File        : in out Stream_IO.File_Type;
      Parent      : in out Page_Type;
      Right_Index : in     Positive;
      Right       : in out Page_Type;
      Underflow   :    out Boolean) is

      -- Description:
      --
      -- Repair an underflow in the Right page, by either moving some extra
      -- items from its left sibling, or by combining the siblings together
      -- to form a single page.
      --
      -- The Underflow flag indicates whether the Parent page has
      -- underflowed.
      --
      -- The Parent page and both of its children are all modified.  This
      -- subprogram writes the children to disk (as necessary), but it does
      -- NOT write the Parent page to disk.


      pragma Assert (Right.Length < Order);

      Left_Index : constant Natural :=
        Right_Index - 1;

      Left_File_Index : constant Stream_IO.Positive_Count :=
        Parent.Children (Left_Index);

      Left : Page_Type;

   begin

      Set_Mode (File, In_File);
      Read (File, Left, Left_File_Index);

      Set_Mode (File, Out_File);

      if Left.Length > Order then

         -- There are some items available in the left sibling.  Move
         -- some of the them into the right sibling.
         --
         Redistribute_Items_From_Left_To_Right
           (Parent      => Parent,
            Left        => Left,
            Right_Index => Right_Index,
            Right       => Right);

         Underflow := False;
         -- Let the caller know that the Parent hasn't underflowed.

         Write (File, Right, Parent.Children (Right_Index));

      else

         -- There aren't any extra items available in the left sibling.
         -- Move all of the items in the right child into the left.
         --
         Concatenate
           (Parent      => Parent,
            Left        => Left,
            Right_Index => Right_Index,
            Right       => Right);

         Underflow := Parent.Length < Order;
         -- Let the caller know whether the Parent has underflowed.

         --
         -- Note that we don't write the Right page to disk, because that
         -- page is gone.  All of its items were moved to the Left page as
         -- a result of the concatentation.
         --

      end if;

      Write (File, Left, Left_File_Index);

   end Redistribute_Or_Concatenate_With_Left_Sibling;


   procedure Redistribute_Or_Concatenate_With_Either_Sibling
     (File         : in out Stream_IO.File_Type;
      Parent       : in out Page_Type;
      Middle_Index : in     Positive;
      Child        : in out Page_Type;
      Underflow    :    out Boolean) is

      -- Description:
      --
      -- Repair an underflow in the Child page (located somewhere in the
      -- middle of its parent, "middle" meaning not at either extreme), by
      -- either moving some extra items from one of its two siblings, or by
      -- combining the siblings (Child and its left sibling) together to
      -- form a single page.
      --
      -- The Underflow flag indicates whether the Parent page has
      -- underflowed.
      --
      -- The Parent page, the Child page, and one of its sibling pages are
      -- all modified.  This subprogram writes the children to disk (as
      -- necessary), but it does NOT write the Parent page to disk.


      pragma Assert (Child.Length < Order);

      Child_File_Index : constant Stream_IO.Positive_Count :=
        Parent.Children (Middle_Index);

      Right_Index : constant Positive :=
        Middle_Index + 1;

      Right_File_Index : constant Stream_IO.Positive_Count :=
        Parent.Children (Right_Index);

      Right : Page_Type;

   begin

      Set_Mode (File, In_File);
      Read (File, Right, Right_File_Index);

      if Right.Length > Order then

         -- The right sibling has some extra items.  Repair the underflow
         -- in Child by moving some items from its Right sibling.
         --
         Redistribute_Items_From_Right_To_Left
           (Parent      => Parent,
            Right_Index => Right_Index,
            Right       => Right,
            Left        => Child);

         Underflow := False;
         -- Let the caller know that the Parent page hasn't underflowed.

         Set_Mode (File, Out_File);
         Write (File, Right, Right_File_Index);
         Write (File, Child, Child_File_Index);

      else

         -- There aren't any extra items in the Child's right sibling, so
         -- repair the underflow by moving some items from the left.
         --
         Redistribute_Or_Concatenate_With_Left_Sibling
           (File        => File,
            Parent      => Parent,
            Right_Index => Middle_Index,
            Right       => Child,
            Underflow   => Underflow);

      end if;

   end Redistribute_Or_Concatenate_With_Either_Sibling;



   procedure Redistribute_Or_Concatenate
     (File              : in out Stream_IO.File_Type;
      Parent_File_Index : in     Stream_IO.Positive_Count;
      Parent            : in out Page_Type;
      Child             : in out Page_Type;
      Index             : in     Natural;
      Underflow         :    out Boolean) is

      -- Description:
      --
      -- Repair an underflow in the Child page, by either moving some extra
      -- items from one of its two siblings, or by combining the siblings
      -- together to form a single page.
      --
      -- The Underflow flag indicates whether the Parent page has
      -- underflowed.
      --
      -- The Parent page, the Child page, and one of its sibling pages are
      -- all modified.  This subprogram writes the children to disk (as
      -- necessary).  It also writes the Parent page to disk, but ONLY if
      -- the parent hasn't underflowed.

   begin

      if Index = 0 then

         -- The left-most child underflowed.  We have no choice but to
         -- redistribute or concatenate with right sibling.

         Redistribute_Or_Concatenate_With_Right_Sibling
           (File       => File,
            Parent     => Parent,
            Left_Index => Index,
            Left       => Child,
            Underflow  => Underflow);

      elsif Index = Parent.Length then

         -- The right-most child has an underflow.  We have no choice but
         -- to redistribute or concatenate with the left sibling.

         Redistribute_Or_Concatenate_With_Left_Sibling
           (File        => File,
            Parent      => Parent,
            Right_Index => Index,
            Right       => Child,
            Underflow   => Underflow);

      else

         pragma Assert (Parent.Length > 1);

         -- Neither the left-most nor right-most child has underflowed.
         -- Try both neighbors to find one that has some extra items.

         Redistribute_Or_Concatenate_With_Either_Sibling
           (File         => File,
            Parent       => Parent,
            Middle_Index => Index,
            Child        => Child,
            Underflow    => Underflow);

      end if;


      if not Underflow then
         Write (File, Parent, Parent_File_Index);
      end if;

   end Redistribute_Or_Concatenate;


   procedure Remove_From_Page
     (File            : in out Stream_IO.File_Type;
      Page_File_Index : in     Stream_IO.Positive_Count;
      Page            : in out Page_Type;
      Index           : in     Natural;
      Underflow       :    out Boolean) is

      -- Description:
      --
      -- Deletes the item Page.Items (Index) from Page.
      --
      -- The Underflow flag indicates whether Page has underflowed as a
      -- result of the deletion.
      --
      -- If the Page hasn't underflowed, then it saves Page to disk.


      procedure Swap
        (Subpage            : in out Page_Type;
         Subpage_File_Index : in     Stream_IO.Positive_Count) is

         -- Description:
         --
         -- Replace Page.Items (Index) with the item having largest key in
         -- the subtree whose root is Subpage.
         --
         -- Modifies Underflow to indicate whether Subpage has underflowed.
         --
         -- If Subpage hasn't underflowed, it saves Subpage to disk.


         Length : Natural renames Subpage.Length;
         Child  : Page_Type;

      begin -- Swap

         if Subpage.Children (0) = 0 then

            -- This is rightmost leaf page in the subtree whose root is
            -- Page.  Move the largest item in the leaf up into the Page.
            --
            Page.Items (Index) := Subpage.Items (Length);

            -- Now delete the item from the leaf.
            Length := Length - 1;

            -- Let caller know whether Subpage has underflowed.
            Underflow := Length < Order;

            if not Underflow then
               Set_Mode (File, Out_File);
               Write (File, Subpage, Subpage_File_Index);
            end if;

         else

            Set_Mode (File, In_File);
            Read (File, Child, Subpage.Children (Length));

            Swap (Child, Subpage.Children (Length));

            if Underflow then

               -- The child has underflowed, so repair the underflow.

               Redistribute_Or_Concatenate
                 (File              => File,
                  Parent_File_Index => Subpage_File_Index,
                  Parent            => Subpage,
                  Child             => Child,
                  Index             => Length,
                  Underflow         => Underflow);

               --
               -- The Underflow flag now indicates whether Subpage has
               -- underflowed (because we had to repair the underflow in
               -- its Child).
               --

            end if;

         end if;

      end Swap;


      Child : Page_Type;


   begin -- Remove_From_Page

      if Page.Children (0) = 0 then

         -- This is a leaf page.  Delete the item from this page.

         Page.Items (Index .. Page.Length - 1) :=
           Page.Items (Index + 1 .. Page.Length);

         Page.Children (Index .. Page.Length - 1) :=
           Page.Children (Index + 1 .. Page.Length);

         Page.Length := Page.Length - 1;


         -- Let caller know whether Page has underflowed.
         Underflow := Page.Length < Order;

         if not Underflow then
            Set_Mode (File, Out_File);
            Write (File, Page, Page_File_Index);
         end if;

      else

         -- In a B-tree, you always delete from a leaf page.  This isn't a
         -- leaf page.  However, the item we want to delete is on this page!
         --
         -- We descend the tree, all the way to leaf, to find the item's
         -- left neighbor (the largest item less than Page.Items (Index)).
         -- It is the last (largest) item in the right-most leaf node in
         -- the left subtree.
         --
         -- We "swap" the two items, and then remove the item from the
         -- leaf.  But we don't really swap: we merely copy the left
         -- neighbor from the leaf into the Index'th slot of this page
         -- (thus "deleting" the item from this page - which is our
         -- ultimate intent), and delete that other item from the leaf.
         --
         -- Of course, deleting the item from the leaf may cause an
         -- underflow.

         Set_Mode (File, In_File);
         Read (File, Child, Page.Children (Index - 1));

         Swap (Child, Page.Children (Index - 1));

         if Underflow then

            -- Repair the Underflow in the Child.
            --
            Redistribute_Or_Concatenate
              (File              => File,
               Parent_File_Index => Page_File_Index,
               Parent            => Page,
               Child             => Child,
               Index             => Index - 1,
               Underflow         => Underflow);

            --
            -- The Underflow flag now indicates whether Page has
            -- underflowed (as a result of repairing its Child).
            --

         else

            -- No, the Child hasn't underflowed, but Page was still
            -- modified (during the Swap, the item in the Index'th slot was
            -- replaced by the largest item in Page's left subtree).
            -- Therefore we need to save it to disk.

--          Set_Mode (File, Out_File);
            Write (File, Page, Page_File_Index);

         end if;

      end if;

  end Remove_From_Page;


   procedure Remove_From_Tree
     (File            : in out Stream_IO.File_Type;
      Page_File_Index : in     Stream_IO.Positive_Count;
      Page            : in out Page_Type;
      Key             : in     Key_Type;
      Underflow       :    out Boolean) is

      -- Description:
      --
      -- Remove the item with Key from the tree whose root is Page.
      --
      -- The Underflow flag indicates whether Page has underflowed (as a
      -- result of the removal).
      --
      -- If the Page has been modified, but hasn't underflowed, then it
      -- writes the page to disk.


      Search_Index : Natural;
      Found        : Boolean;

      Child : Page_Type;

   begin

      Search
        (Items => Page.Items (1 .. Page.Length),
         Key   => Key,
         Index => Search_Index,
         Found => Found);


      if Found then

         Remove_From_Page
           (File            => File,
            Page_File_Index => Page_File_Index,
            Page            => Page,
            Index           => Search_Index,
            Underflow       => Underflow);

      elsif Page.Children (0) = 0 then

         -- This is a leaf, so we descended the entire tree without finding
         -- the item.  Therefore, we're done.

         -- Let caller know that Page hasn't underflowed.
         Underflow := False;

      else

         -- The item isn't on this page.  Keep descending the tree to find
         -- the item.

--       Set_Mode (File, In_File);
         Read (File, Child, Page.Children (Search_Index));

         Remove_From_Tree
           (File            => File,
            Page_File_Index => Page.Children (Search_Index),
            Page            => Child,
            Key             => Key,
            Underflow       => Underflow);

         if Underflow then

            -- Repair the underflow in the Child page.
            --
            Redistribute_Or_Concatenate
              (File              => File,
               Parent_File_Index => Page_File_Index,
               Parent            => Page,
               Child             => Child,
               Index             => Search_Index,
               Underflow         => Underflow);

            --
            -- Note the the Underflow flag now indicates whether an
            -- underflow has occured in Page (as a result of repairing the
            -- underflow in its Child.)
            --

         end if;

      end if;

   end Remove_From_Tree;


   procedure Search
     (Items : in     Item_Array_Base;
      Key   : in     Key_Type;
      Index :    out Natural;
      Found :    out Boolean) is

      -- Description:
      --
      -- Performs a binary search of Items.
      --
      -- If Key is found, then Found is True and Index is the array index
      -- of the matching item.
      --
      -- If Key is not found, then Found is False and Index is the array
      -- index of the largest key less than Key.
      --
      -- If the length of the Items array is 0, then Found is False and
      -- Index is 0.


      Lower : Natural := 0;
      Upper : Positive := Items'Length + 1;

      Middle : Positive range 1 .. Items'Length;

   begin

      while Upper - Lower > 1 loop

         Middle := (Lower + Upper) / 2;

         declare
            Middle_Key : Key_Type renames Get_Key (Items (Middle));
         begin
            if Middle_Key < Key then

               Lower := Middle;

            elsif Middle_Key = Key then

               Index := Middle;
               Found := True;

               return;

            else

               Upper := Middle;

            end if;
         end;

      end loop;

      Index := Lower;
      Found := False;

   end Search;

end Indexed_IO;


