------------------------------------------------------------------------
--
--                               INDEXED_IO
--
--      $Id: indexed_io.ads,v 1.10.1.20 1999/10/12 00:09:31 matt Exp $
--
-- Copyright (C) 1999 Matthew J Heaney <matthew_heaney@acm.org>
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or (at
-- your option) any later version.
--
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
-- USA
--
-- As a special exception, if other files instantiate this generic unit,
-- or you link this unit with other files to produce an executable, this
-- unit does not by itself cause the resulting executable to be covered
-- by the GNU General Public License.  This exception does not however
-- invalidate any other reasons why the executable file might be covered
-- by the GNU Public License.
--
------------------------------------------------------------------------

with Ada.Streams.Stream_IO;
with Ada.IO_Exceptions;

generic

   type Item_Type is private;

   type Key_Type (<>) is limited private;

   Order : in Positive;
   -- Defined as the minimum number of items per page.

   with function Get_Key (Item : Item_Type) return Key_Type is <>;

   with function "=" (L, R : Key_Type) return Boolean is <>;
   with function "<" (L, R : Key_Type) return Boolean is <>;

package Indexed_IO is

   pragma Elaborate_Body;


   type File_Type is limited private;


   procedure Create
     (File : in out File_Type;
      Name : in     String := "");

   procedure Open
     (File : in out File_Type;
      Name : in     String);

   procedure Close
     (File : in out File_Type);

   function Is_Open
     (File : File_Type) return Boolean;


   procedure Read
     (File : in     File_Type;
      Key  : in     Key_Type;
      Item :    out Item_Type);

   procedure Write
     (File : in File_Type;
      Item : in Item_Type);

   procedure Remove
     (File : in File_Type;
      Key  : in Key_Type);


   generic

      with procedure Process
        (Item : in     Item_Type;
         Done : in out Boolean) is <>;

   procedure Select_Items
     (File : in File_Type);


   generic

      with procedure Process
        (Item : in     Item_Type;
         Done : in out Boolean) is <>;

   procedure Select_Items_Starting_At_Key
     (File : in File_Type;
      Key  : in Key_Type);



   Item_Not_Found : exception;
   Status_Error   : exception renames Ada.IO_Exceptions.Status_Error;
   Mode_Error     : exception renames Ada.IO_Exceptions.Mode_Error;
   Name_Error     : exception renames Ada.IO_Exceptions.Name_Error;
   Use_Error      : exception renames Ada.IO_Exceptions.Use_Error;
   Device_Error   : exception renames Ada.IO_Exceptions.Device_Error;
   End_Error      : exception renames Ada.IO_Exceptions.End_Error;
   Data_Error     : exception renames Ada.IO_Exceptions.Data_Error;

private

   use Ada.Streams;


   Page_Size : constant Positive := 2 * Order;
   -- Defined as the maximum number of items per page.

   subtype Item_Range is Positive range 1 .. Page_Size;

   type Item_Array_Base is
      array (Positive range <>) of Item_Type;


   subtype Child_Range is
     Natural range 0 .. Page_Size;

   type File_Index_Array is
      array (Child_Range) of Stream_IO.Count;

   type Page_Type is
      limited record
         Length   : Natural;
         Items    : Item_Array_Base (Item_Range);
         Children : File_Index_Array;
      end record;


   type Handle_Type (File : access File_Type) is
      limited null record;

   type File_Type is
      limited record
         Handle               : Handle_Type (File_Type'Access);
         File                 : Stream_IO.File_Type;
         Root_Page_File_Index : Stream_IO.Positive_Count;
         Root_Page            : Page_Type;
      end record;


   procedure Read
     (File : in     Stream_IO.File_Type;
      Page :    out Page_Type;
      From : in     Stream_IO.Positive_Count);


   procedure Search
     (Items : in     Item_Array_Base;
      Key   : in     Key_Type;
      Index :    out Natural;
      Found :    out Boolean);

end Indexed_IO;
